-- BIT_XOR
-- Note: only works for integers and bigints
CREATE OR REPLACE FUNCTION _bit_xor(bigint, bigint)
RETURNS bigint AS $$
  SELECT $1 # COALESCE($2, 0)
$$ IMMUTABLE LANGUAGE SQL;

CREATE AGGREGATE bit_xor (
	BASETYPE = bigint,
	SFUNC = _bit_xor,
	STYPE = bigint,
	INITCOND = 0
);

-- GROUP_CONCAT()
-- Note: only supports the comma separator
-- Note: For DISTINCT and ORDER BY a subquery is required
CREATE OR REPLACE FUNCTION _group_concat(text, text)
RETURNS text AS $$
  SELECT CASE
    WHEN $2 IS NULL THEN $1
    WHEN $1 IS NULL THEN $2
    ELSE $1 operator(pg_catalog.||) ',' operator(pg_catalog.||) $2
  END
$$ IMMUTABLE LANGUAGE SQL;

CREATE AGGREGATE group_concat (
	BASETYPE = text,
	SFUNC = _group_concat,
	STYPE = text
);

-- IF
-- Warning: still requires casts in some instances
CREATE OR REPLACE FUNCTION if(boolean, anyelement, anyelement) 
RETURNS anyelement AS '
  SELECT CASE WHEN $1 THEN $2 ELSE $3 END
' IMMUTABLE LANGUAGE SQL;

-- IFNULL
-- Warning: still requires casts in some instances
CREATE OR REPLACE FUNCTION ifnull(anyelement, anyelement)
RETURNS anyelement AS '
  SELECT COALESCE($1, $2)
' IMMUTABLE LANGUAGE SQL;

-- BENCHMARK()
-- Note: This version requires the expression to be quoted.
-- Note: To show query times in psql, run \timing first
-- Example: SELECT BENCHMARK(100000, $$ length('hello') $$);
CREATE OR REPLACE FUNCTION benchmark(integer, text)
RETURNS integer AS $$
  BEGIN
    FOR i IN 1..$1 LOOP
      EXECUTE 'SELECT ' || $2;
    END LOOP;
    RETURN 0;
  END;
$$ STRICT LANGUAGE PLPGSQL;

-- CHARSET()
-- This is a bit dodgy as it just returns the database encoding
CREATE OR REPLACE FUNCTION charset(text)
RETURNS text AS $$
  SELECT pg_catalog.lower(setting) from pg_catalog.pg_settings where name='server_encoding'
$$ IMMUTABLE LANGUAGE SQL;

-- COERCIBILITY()
-- This is a bit dodgy as PostgreSQL does not support collations
-- Note: This is MySQL 5.0 compatible
CREATE OR REPLACE FUNCTION coercibility(name)
RETURNS integer AS $$
  SELECT 3
$$ IMMUTABLE LANGUAGE SQL;

CREATE OR REPLACE FUNCTION coercibility(text)
RETURNS integer AS $$
  SELECT CASE
    WHEN $1 IS NULL THEN 5
    ELSE 2
  END
$$ IMMUTABLE LANGUAGE SQL;

-- COLLATION()
-- In PostgreSQL 9.1 and earlier, just return the database collation
-- In 9.2 COLLATION is a reserved keyword, so this function cannot provide
-- compatbility anyway. Please use COLLATION FOR(x) instead of COLLATION(x)
--
-- Not sure why this calls lower() -- collation names are case sensitive in Postgres
CREATE OR REPLACE FUNCTION "collation"(text)
RETURNS text AS $$
  SELECT pg_catalog.lower(setting) from pg_catalog.pg_settings where name='lc_collate'
$$ IMMUTABLE LANGUAGE SQL;

-- CONNECTION_ID()
CREATE OR REPLACE FUNCTION connection_id()
RETURNS integer AS $$ 
  SELECT pg_catalog.pg_backend_pid()
$$ IMMUTABLE LANGUAGE SQL;

-- DATABASE()
CREATE OR REPLACE FUNCTION database()
RETURNS text AS $$
  SELECT pg_catalog.current_database()::text
$$ IMMUTABLE LANGUAGE SQL;

-- FOUND_ROWS()
-- Not possible to implement

-- LAST_INSERT_ID()
-- Note: only works in 8.1
-- Note: Not possible to implement last_insert_id(val)
CREATE OR REPLACE FUNCTION last_insert_id()
RETURNS bigint AS $$
  SELECT pg_catalog.lastval()
$$ VOLATILE LANGUAGE SQL;

-- ROW_COUNT()
-- Not possible to implement

-- SCHEMA()
-- XXX: Should this be an alias to database() like in mysql?
CREATE OR REPLACE FUNCTION schema()
RETURNS text AS $$
  SELECT pg_catalog.current_schema()::text
$$ VOLATILE LANGUAGE SQL;

-- SESSION_USER()
-- Cannot be implemented

-- SYSTEM_USER()
CREATE OR REPLACE FUNCTION system_user()
RETURNS text AS $$
  SELECT SESSION_USER::text
$$ VOLATILE LANGUAGE SQL;

-- USER()
-- Cannot be implemented
-- TIME TO INTEGER
CREATE OR REPLACE FUNCTION _time_to_integer(time with time zone)
RETURNS integer AS $$
  SELECT
    EXTRACT(HOUR FROM $1)::integer * 10000
    + EXTRACT(MINUTE FROM $1)::integer * 100
    + EXTRACT(SECONDS FROM $1)::integer
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE CAST (time with time zone AS integer)
  WITH FUNCTION _time_to_integer (time with time zone)
  AS IMPLICIT;

CREATE OR REPLACE FUNCTION _time_to_integer(time without time zone)
RETURNS integer AS $$
  SELECT
    EXTRACT(HOUR FROM $1)::integer * 10000
    + EXTRACT(MINUTE FROM $1)::integer * 100
    + EXTRACT(SECONDS FROM $1)::integer
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE CAST (time without time zone AS integer)
  WITH FUNCTION _time_to_integer (time without time zone)
  AS IMPLICIT;

-- DATE TO INTEGER
-- Note: requires explicit casts in some cases
CREATE OR REPLACE FUNCTION _date_to_integer(date)
RETURNS integer AS $$
  SELECT
    EXTRACT(YEAR FROM $1)::integer * 10000
    + EXTRACT(MONTH FROM $1)::integer * 100
    + EXTRACT(DAY FROM $1)::integer
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE CAST (date AS integer)
  WITH FUNCTION _date_to_integer (date)
  AS IMPLICIT;

-- TIMESTAMP TO BIGINT
-- Depends on: _time_to_integer and _date_to_integer
CREATE OR REPLACE FUNCTION _timestamp_to_bigint(timestamp with time zone)
RETURNS bigint AS $$
  SELECT _date_to_integer($1::date)::bigint * 1000000 + _time_to_integer($1::time)
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE CAST (timestamp with time zone AS bigint)
  WITH FUNCTION _timestamp_to_bigint (timestamp with time zone)
  AS IMPLICIT;

CREATE OR REPLACE FUNCTION _timestamp_to_bigint(timestamp without time zone)
RETURNS bigint AS $$
  SELECT _date_to_integer($1::date)::bigint * 1000000 + _time_to_integer($1::time)
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE CAST (timestamp without time zone AS bigint)
  WITH FUNCTION _timestamp_to_bigint (timestamp without time zone)
  AS IMPLICIT;

-- INTERVAL TO BIGINT
CREATE OR REPLACE FUNCTION _interval_to_bigint(interval)
RETURNS bigint AS $$
  SELECT
    EXTRACT(YEAR FROM $1)::bigint * 10000000000
    + EXTRACT(MONTH FROM $1)::bigint * 100000000
    + EXTRACT(DAY FROM $1)::bigint * 1000000
    + EXTRACT(HOUR FROM $1)::bigint * 10000
    + EXTRACT(MINUTE FROM $1)::bigint * 100
    + EXTRACT(SECONDS FROM $1)::bigint
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE CAST (interval AS bigint)
  WITH FUNCTION _interval_to_bigint (interval)
  AS IMPLICIT;

-- BIT_COUNT()
CREATE OR REPLACE FUNCTION bit_count(bigint)
RETURNS integer AS $$
  SELECT pg_catalog.length(pg_catalog.replace(pg_catalog.ltrim(pg_catalog.textin(pg_catalog.bit_out($1::bit(64))), '0'), '0', ''));
$$ IMMUTABLE STRICT LANGUAGE SQL;
-- BIN()
CREATE OR REPLACE FUNCTION bin(bigint)
RETURNS text AS $$ 
  SELECT pg_catalog.ltrim(pg_catalog.textin(pg_catalog.bit_out($1::bit(64))), '0');
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- CHAR()
-- Not implemented

-- COMPRESS
-- Not implemented

-- CONCAT
CREATE OR REPLACE FUNCTION concat(text)
RETURNS text AS $$
  SELECT $1
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION concat(text, text)
RETURNS text AS $$
  SELECT $1 operator(pg_catalog.||) $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION concat(text, text, text)
RETURNS text AS $$
  SELECT $1 operator(pg_catalog.||) $2 operator(pg_catalog.||) $3
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- CONCAT_WS
-- Note: fails in this case: select concat_ws(',', 'First name', null);
CREATE OR REPLACE FUNCTION concat_ws(text, text)
RETURNS text AS $$
  SELECT CASE
    WHEN $1 IS NULL THEN NULL
    ELSE $2
  END
$$ IMMUTABLE LANGUAGE SQL;

CREATE OR REPLACE FUNCTION concat_ws(text, text, text)
RETURNS text AS $$
  SELECT CASE
    WHEN $1 IS NULL THEN NULL
    ELSE
      coalesce($2 operator(pg_catalog.||) $1, '') operator(pg_catalog.||) coalesce($3, '')
  END
$$ IMMUTABLE LANGUAGE SQL;

CREATE OR REPLACE FUNCTION concat_ws(text, text, text, text)
RETURNS text AS $$
  SELECT CASE
    WHEN $1 IS NULL THEN NULL
    ELSE 
      coalesce($2 operator(pg_catalog.||) $1, '') operator(pg_catalog.||) coalesce($3 operator(pg_catalog.||) $1, '') operator(pg_catalog.||) coalesce($4, '')
  END
$$ IMMUTABLE LANGUAGE SQL;

-- CONV()
-- Credit: Gavin Sherry
create or replace function _todec(text, int)
returns int as $$
declare
    num int := 0;
    hex text := '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';
    pos int := 0;
    chr text;
    numin alias for $1;
    base alias for $2;
begin
    if numin isnull or base isnull then
        return null;
    end if;
    for i in 1 .. pg_catalog.length(numin) loop
        pos := pg_catalog.abs(position(pg_catalog.upper(substring(numin from i for 1)) in hex) - 1);
        num := num * base + pos;
    end loop;
    return num;
end$$
language plpgsql
RETURNS NULL ON NULL INPUT
IMMUTABLE;

create or replace function conv(int, int, int)
returns text as
$$
declare
    res text := '';
    hex text := '0123456789ABCDEFGHIJLMNOPQRSTUVWXYZ';
    num int;
    tmp int;
    tmp2 text;
    numin text;
    tobase int;
    isneg bool := false;
    numin_p ALIAS FOR $1;
    frombase ALIAS FOR $2;
    tobase_p ALIAS FOR $3;
begin
    if numin_p < 0 and tobase_p < 0 then
      isneg := true;
      numin := numin_p::integer * -1;
      tobase := tobase_p * -1;
    else
      numin := numin_p;
      tobase := tobase_p;
    end if;

    if numin isnull OR frombase isnull OR tobase ISNULL then
        return NULL;
    elsif frombase < 0 OR frombase > 36 then
        return NULL;
    elsif tobase < 0 OR tobase > 36 then
        return NULL;
    end if;
 
    if frombase <> 10 then
        num := _todec(numin, frombase);
    else
        num := numin::int;
    end if;
 
    loop
        tmp := num % tobase + 1;
        res := substring( hex from tmp for 1 ) operator(pg_catalog.||) res;
        num := num/tobase;
        if num = 0 then
            exit;
        end if;
    end loop;
    if isneg then
        return '-' operator(pg_catalog.||) res;
    else
        return res;
    end if;
end
$$
language plpgsql
RETURNS NULL ON NULL INPUT
IMMUTABLE;
 
create or replace function conv(text, int, int)
returns text as
$$
declare
    res text := '';
    hex text := '0123456789ABCDEFGHIJLMNOPQRSTUVWXYZ';
    num int;
    tmp int;
    tmp2 text;
    numin text;
    tobase int;
    isneg bool := false;
    numin_p ALIAS FOR $1;
    frombase ALIAS FOR $2;
    tobase_p ALIAS FOR $3;
begin
    if numin_p < '0' and tobase_p < 0 then
      isneg := true;
      numin := numin_p::integer * -1;
      tobase := tobase_p * -1;
    else
      numin := numin_p;
      tobase := tobase_p;
    end if;

    if numin isnull OR frombase isnull OR tobase ISNULL then
        return NULL;
    elsif frombase < 0 OR frombase > 36 then
        return NULL;
    elsif tobase < 0 OR tobase > 36 then
        return NULL;
    end if;
 
    if frombase <> 10 then
        num := _todec(numin, frombase);
    else
        num := numin::int;
    end if;
 
    loop
        tmp := num % tobase + 1;
        res := substring( hex from tmp for 1 ) operator(pg_catalog.||) res;
        num := num/tobase;
        if num = 0 then
            exit;
        end if;
    end loop;
    if isneg then
        return '-' operator(pg_catalog.||) res;
    else
        return res;
    end if;
end
$$
language plpgsql
RETURNS NULL ON NULL INPUT
IMMUTABLE;
 
-- ELT()
CREATE OR REPLACE FUNCTION elt(integer, text, text)
RETURNS text AS $$
  SELECT CASE
    WHEN $1 < 1 OR $1 > 2 THEN NULL
    WHEN $1 = 1 THEN $2
    ELSE $3
  END 
$$ IMMUTABLE LANGUAGE SQL;

CREATE OR REPLACE FUNCTION elt(integer, text, text, text)
RETURNS text AS $$
  SELECT CASE
    WHEN $1 < 1 OR $1 > 3 THEN NULL
    WHEN $1 = 1 THEN $2
    WHEN $1 = 2 THEN $3
    ELSE $4
  END 
$$ IMMUTABLE LANGUAGE SQL;

CREATE OR REPLACE FUNCTION elt(integer, text, text, text, text)
RETURNS text AS $$
  SELECT CASE
    WHEN $1 < 1 OR $1 > 4 THEN NULL
    WHEN $1 = 1 THEN $2
    WHEN $1 = 2 THEN $3
    WHEN $1 = 3 THEN $4
    ELSE $5
  END 
$$ IMMUTABLE LANGUAGE SQL;

-- REVERSE()
CREATE OR REPLACE FUNCTION reverse(text)
RETURNS text AS $$
  DECLARE
    temp TEXT;
    count INTEGER;
  BEGIN
    temp := '';
    count := pg_catalog.length($1);
    FOR i IN REVERSE count..1 LOOP
      temp := temp  operator(pg_catalog.||)  substring($1 from i for 1);
    END LOOP;
    RETURN temp;
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- EXPORT_SET()
-- Depends on: BIN() and REVERSE()
-- XXX: WILL fail if $2 is '0'
CREATE OR REPLACE FUNCTION export_set(bigint, text, text, text, integer)
RETURNS text AS $$
  SELECT pg_catalog.rtrim(pg_catalog.replace(pg_catalog.replace(reverse(pg_catalog.lpad(bin($1), $5, '0')), '1', $2 operator(pg_catalog.||) $4), '0', $3 operator(pg_catalog.||) $4), $4)
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION export_set(bigint, text, text, text)
RETURNS text AS $$
  SELECT export_set($1, $2, $3, $4, 64)
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION export_set(bigint, text, text)
RETURNS text AS $$
  SELECT export_set($1, $2, $3, ',', 64)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- FIELD()
CREATE OR REPLACE FUNCTION field(anyelement , VARIADIC anyarray ) RETURNS INTEGER AS $$
	SELECT i
	  FROM generate_subscripts($2, 1) AS i
	 WHERE $2[i] IS NOT DISTINCT FROM $1
$$ LANGUAGE SQL;

CREATE OR REPLACE FUNCTION field(bigint, VARIADIC anyarray ) RETURNS INTEGER AS $$
	SELECT i
	  FROM generate_subscripts($2, 1) AS i
	 WHERE $2[i] IS NOT DISTINCT FROM $1
$$ LANGUAGE SQL;

-- FIND_IN_SET()
CREATE OR REPLACE FUNCTION find_in_set(text, text)
RETURNS integer AS $$
  DECLARE
    list text[];
    len integer;
  BEGIN
    IF $2 = '' THEN
      RETURN 0;
    END IF;
    list := pg_catalog.string_to_array($2, ',');
    len := pg_catalog.array_upper(list, 1);
    FOR i IN 1..len LOOP
      IF list[i] = $1 THEN
        RETURN i;
      END IF;
    END LOOP;
    RETURN 0;
  END;
$$ STRICT IMMUTABLE LANGUAGE PLPGSQL;

-- HEX()
CREATE OR REPLACE FUNCTION hex(integer)
RETURNS text AS $$
  SELECT pg_catalog.upper(pg_catalog.to_hex($1))
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION hex(bigint)
RETURNS text AS $$
  SELECT pg_catalog.upper(pg_catalog.to_hex($1))
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION hex(text)
RETURNS text AS $$
  DECLARE
    len integer;
    temp text;
  BEGIN
    len := pg_catalog.length($1);
    temp := '';
    FOR i IN 1..len LOOP
      temp := temp operator(pg_catalog.||) pg_catalog.to_hex(pg_catalog.ascii(SUBSTRING($1 FROM i FOR 1)));
    END LOOP;
    RETURN pg_catalog.upper(temp);
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- FORMAT()
-- See: mathematical.sql

-- INSERT()
CREATE OR REPLACE FUNCTION insert(text, integer, integer, text)
RETURNS text AS $$
  SELECT CASE
    WHEN NOT $2 BETWEEN 1 AND pg_catalog.length($1) THEN $1
    ELSE overlay($1 placing $4 from $2 for $3)
  END
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- INSTR()
CREATE OR REPLACE FUNCTION instr(text, text)
RETURNS integer AS $$
  SELECT POSITION($2 IN $1)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- LCASE()
CREATE OR REPLACE FUNCTION lcase(text)
RETURNS text AS $$
  SELECT pg_catalog.lower($1)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- LEFT()
CREATE OR REPLACE FUNCTION left(text, integer)
RETURNS text AS $$
  SELECT substring($1 FOR $2);
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- LOAD_FILE()
-- Not implemented

-- LOCATE()
CREATE OR REPLACE FUNCTION locate(text, text, integer)
RETURNS integer AS $$
  SELECT POSITION($1 IN SUBSTRING ($2 FROM $3)) + $3 - 1
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION locate(text, text)
RETURNS integer AS $$
  SELECT locate($1, $2, 1)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- MAKE_SET()
-- routine to make make_set() easier
-- note: since arrays do not support NULLs until 8.2, we cannot
-- handle NULL arguments to make_set()
-- note: we only support 32 arguments to make_set() as PostgreSQL
-- does not support arbitary argument lists
CREATE OR REPLACE FUNCTION _make_set(bigint, text[])
RETURNS text AS $$
  DECLARE
    up int;
    i int = 1;
    ret text = '';
    dat ALIAS FOR $2;
    found bool = false;
    mask bigint = 1;
  BEGIN
    up = array_upper($2, 1);
    IF up > 31 THEN
      RAISE EXCEPTION 'maximum array size exceeded';
      RETURN NULL;
    END IF;
    WHILE i <= up LOOP
      IF $1 & mask <> 0 THEN
        IF found = false THEN
          found = true;
        ELSE
          ret = ret operator(pg_catalog.||) ',';
        END IF;
        ret = ret operator(pg_catalog.||) dat[i];
      END IF;
      i = i + 1;
      mask = mask << 1;
    END LOOP;
    RETURN ret;
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;


CREATE OR REPLACE FUNCTION make_set(bigint, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28, $29];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28, $29, $30];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28, $29, $30, $31];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION make_set(bigint, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text, text)
RETURNS text AS $$
  DECLARE
    a text[];
  BEGIN
    a = ARRAY[$2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28, $29, $30, $31, $32];
RETURN _make_set($1, a);
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- MID()
CREATE OR REPLACE FUNCTION mid(text, integer, integer)
RETURNS text AS $$
  SELECT pg_catalog.substring($1, $2, $3)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- OCT()
-- Depends on: CONV()
CREATE OR REPLACE FUNCTION oct(integer)
RETURNS text AS $$
  SELECT conv($1, 10, 8)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- OCTET_LENGTH()
-- not reimplemented

-- ORD()
-- Note: Does not support multibyte
CREATE OR REPLACE FUNCTION ord(text)
RETURNS integer AS $$
  SELECT pg_catalog.ascii($1)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- QUOTE()
CREATE OR REPLACE FUNCTION quote(text)
RETURNS text AS $$
  SELECT CASE
    WHEN $1 IS NULL THEN 'NULL'
    ELSE pg_catalog.quote_literal($1)
  END
$$ IMMUTABLE LANGUAGE SQL;

-- REVERSE()
-- See above.  Needed by EXPORT_SET().

-- RIGHT()
CREATE OR REPLACE FUNCTION right(text, integer)
RETURNS text AS $$
  SELECT substring($1 FROM pg_catalog.length($1) + 1 - $2);
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- SOUNDEX()
-- Thanks to Fredrik Olsson for the original soundex() function.
CREATE OR REPLACE FUNCTION _soundexcode(char(1))
RETURNS char(1) AS $$
  SELECT COALESCE(
    (ARRAY['0', '1', '2', '3', '0', 
           '1', '2', '0', '0', '2', 
           '2', '4', '5', '5', '0', 
           '1', '2', '6', '2', '3', 
           '0', '1', '0', '2', '0', '2'])[pg_catalog.ascii($1) - 64],
     '0');
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION soundex(text) RETURNS text AS $$
DECLARE
        a_text alias for $1;
	l_text text;
	l_lchr char(1);
	l_chr char(1);
	l_ret text;
BEGIN
	l_text := pg_catalog.upper(trim(both from a_text));
	IF l_text = '' THEN
		RETURN '0000';
	END IF;
	l_chr := substring(l_text FOR 1);
	l_ret := l_chr;
	l_text := substring(l_text FROM 2);
	WHILE (l_text <> '') LOOP
		l_lchr := l_chr; 
		l_chr := substring(l_text FOR 1);
		l_text := substring(l_text FROM 2);
		IF (pg_catalog.ascii(l_chr) BETWEEN 65 AND 90) AND
				(_soundexcode(l_chr) <> _soundexcode(l_lchr)) THEN
			IF _soundexcode(l_chr) <> '0' THEN
				l_ret := l_ret operator(pg_catalog.||) _soundexcode(l_chr);
			END IF;
		END IF;
	END LOOP;
	IF pg_catalog.length(l_ret) < 4 THEN
	 	l_ret := rpad(l_ret, 4, '0');
	END IF;
	RETURN l_ret;
END;
$$ IMMUTABLE STRICT LANGUAGE plpgsql;

-- SPACE()
CREATE OR REPLACE FUNCTION space(integer)
RETURNS text AS $$
  SELECT pg_catalog.repeat(' ', $1)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- SUBSTRING_INDEX()
CREATE OR REPLACE FUNCTION substring_index(text, text, integer)
RETURNS text AS $$
  DECLARE
    tokens text[];
  BEGIN
    tokens := pg_catalog.string_to_array($1, $2); 

    IF $3 >= 0 THEN
      RETURN pg_catalog.array_to_string(tokens[1:$3], $2);
    ELSE
      RETURN pg_catalog.array_to_string(tokens[($3 * -1):pg_catalog.array_upper(tokens, 1)], $2);
    END IF;
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- STRCMP()
-- Note: comparison is case-sensitive
CREATE OR REPLACE FUNCTION strcmp(text, text)
RETURNS integer AS $$
  SELECT CASE
    WHEN $1 = $2 THEN 0
    WHEN $1 < $2 THEN -1
    ELSE 1
  END
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- UCASE()
CREATE OR REPLACE FUNCTION ucase(text)
RETURNS text AS $$
  SELECT pg_catalog.upper($1)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- UNCOMPRESS()
-- Not implemented.

-- UNHEX()
-- Depends on: CONV()
CREATE OR REPLACE FUNCTION unhex(text)
RETURNS text AS $$
  DECLARE
    len integer := pg_catalog.length($1);
    temp text := '';
    i int := 1;
  BEGIN
    WHILE i <= len LOOP
      temp := temp operator(pg_catalog.||) pg_catalog.chr(conv(substring($1 from i for 2), 16, 10)::integer);
      i := i + 2;
    END LOOP;
    RETURN temp; 
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;
-- ATAN2()
-- Two parameter alias of atan2
CREATE OR REPLACE FUNCTION atan(double precision, double precision) 
RETURNS double precision AS $$
  SELECT pg_catalog.atan2($1, $2)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- CRC32()
CREATE OR REPLACE FUNCTION crc32(text)
RETURNS bigint AS $$
  DECLARE
    -- crc look up table
    -- based on the polynomial:
    -- x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x+1

    crc_lu bigint[256] := ARRAY[0, 1996959894, 3993919788, 2567524794, 
      124634137, 1886057615, 3915621685, 2657392035, 249268274, 2044508324, 
      3772115230, 2547177864, 162941995, 2125561021, 3887607047, 2428444049,
      498536548, 1789927666, 4089016648, 2227061214, 450548861, 1843258603,
      4107580753, 2211677639, 325883990, 1684777152, 4251122042, 
      2321926636, 335633487, 1661365465, 4195302755, 2366115317, 
      997073096, 1281953886, 3579855332, 2724688242, 1006888145, 
      1258607687, 3524101629, 2768942443, 901097722, 1119000684, 
      3686517206, 2898065728, 853044451, 1172266101, 3705015759, 
      2882616665, 651767980, 1373503546, 3369554304, 3218104598, 
      565507253, 1454621731, 3485111705, 3099436303, 671266974, 
      1594198024, 3322730930, 2970347812, 795835527, 1483230225, 
      3244367275, 3060149565, 1994146192, 31158534, 2563907772, 
      4023717930, 1907459465, 112637215, 2680153253, 3904427059, 
      2013776290, 251722036, 2517215374, 3775830040, 2137656763, 
      141376813, 2439277719, 3865271297, 1802195444, 476864866, 
      2238001368, 4066508878, 1812370925, 453092731, 2181625025, 
      4111451223, 1706088902, 314042704, 2344532202, 4240017532, 
      1658658271, 366619977, 2362670323, 4224994405, 1303535960, 
      984961486, 2747007092, 3569037538, 1256170817, 1037604311, 
      2765210733, 3554079995, 1131014506, 879679996, 2909243462, 
      3663771856, 1141124467, 855842277, 2852801631, 3708648649, 
      1342533948, 654459306, 3188396048, 3373015174, 1466479909, 
      544179635, 3110523913, 3462522015, 1591671054, 702138776, 
      2966460450, 3352799412, 1504918807, 783551873, 3082640443, 
      3233442989, 3988292384, 2596254646, 62317068, 1957810842, 
      3939845945, 2647816111, 81470997, 1943803523, 3814918930, 
      2489596804, 225274430, 2053790376, 3826175755, 2466906013, 
      167816743, 2097651377, 4027552580, 2265490386, 503444072, 
      1762050814, 4150417245, 2154129355, 426522225, 1852507879, 
      4275313526, 2312317920, 282753626, 1742555852, 4189708143, 
      2394877945, 397917763, 1622183637, 3604390888, 2714866558, 
      953729732, 1340076626, 3518719985, 2797360999, 1068828381, 
      1219638859, 3624741850, 2936675148, 906185462, 1090812512, 
      3747672003, 2825379669, 829329135, 1181335161, 3412177804, 
      3160834842, 628085408, 1382605366, 3423369109, 3138078467, 
      570562233, 1426400815, 3317316542, 2998733608, 733239954, 
      1555261956, 3268935591, 3050360625, 752459403, 1541320221, 
      2607071920, 3965973030, 1969922972, 40735498, 2617837225, 
      3943577151, 1913087877, 83908371, 2512341634, 3803740692, 
      2075208622, 213261112, 2463272603, 3855990285, 2094854071, 
      198958881, 2262029012, 4057260610, 1759359992, 534414190, 
      2176718541, 4139329115, 1873836001, 414664567, 2282248934, 
      4279200368, 1711684554, 285281116, 2405801727, 4167216745, 
      1634467795, 376229701, 2685067896, 3608007406, 1308918612, 
      956543938, 2808555105, 3495958263, 1231636301, 1047427035, 
      2932959818, 3654703836, 1088359270, 936918000, 2847714899, 
      3736837829, 1202900863, 817233897, 3183342108, 3401237130, 
      1404277552, 615818150, 3134207493, 3453421203, 1423857449, 
      601450431, 3009837614, 3294710456, 1567103746, 711928724, 
      3020668471, 3272380065, 1510334235, 755167117];
    crc bigint := 4294967295; -- must use bigint, because we don't have
                              -- an unsigned int32
    len int;
    data ALIAS FOR $1;
    tmp bigint;
    pos int := 1;
    c text;
    d bigint;
  BEGIN
    len = pg_catalog.length(data);
    WHILE len >= pos LOOP
        c := SUBSTRING(data FROM pos FOR 1);
        d := pg_catalog.ascii(c);
        pos = pos + 1;
        tmp = (crc # d) & 255;
        crc = crc_lu[tmp + 1] # (crc >> 8);
    END LOOP;
    crc := crc # 4294967295;
    RETURN crc;
  END
$$ IMMUTABLE STRICT LANGUAGE plpgsql;

-- FORMAT()
-- FIXME: Only handles numbers with integer part up to 21 digits
--
-- MySQL has some bugs with long numbers, but they're not worth replicating:
-- format(9999999999999999999999999999999999999999999999999999999999999999, 10)
--  --> last group has four nines!
-- format(9999999999999999999999999999999999999999999999999999999999999999999999999, 1);
--  --> last group has five nines and no decmial zero
CREATE OR REPLACE FUNCTION format(numeric, integer)
RETURNS text AS $$
  SELECT pg_catalog.to_char(pg_catalog.round($1, $2), 'FM999,999,999,999,999,999,990'
    operator(pg_catalog.||)
      case when $2 > 0 then
        '.' operator(pg_catalog.||) pg_catalog.repeat('0', $2)
      else '' end)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- LN(), LOG()
-- Not reimplemented but note that PostgreSQL has an error on -ve values,
-- but MySQL just returns NULL.

-- LOG2()
CREATE OR REPLACE FUNCTION log2(numeric)
RETURNS numeric AS $$
  SELECT CASE WHEN $1 > 0 THEN pg_catalog.log(2, $1) ELSE NULL END
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- LOG10()
CREATE OR REPLACE FUNCTION log10(numeric)
RETURNS numeric AS $$
  SELECT CASE WHEN $1 > 0 THEN pg_catalog.log(10, $1) ELSE NULL END
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- MOD()
-- PostgreSQL has all MOD usage the same as MySQL EXCEPT this will not work:
--
-- SELECT 29 MOD 9;

-- RAND()
CREATE OR REPLACE FUNCTION rand() RETURNS double precision AS $$
  SELECT pg_catalog.random()
$$ VOLATILE LANGUAGE SQL;

CREATE OR REPLACE FUNCTION rand(integer) RETURNS double precision AS $$
  SELECT pg_catalog.setseed($1); 
  SELECT pg_catalog.random()
$$ VOLATILE LANGUAGE SQL;

-- SQRT()
-- Not reimplemented but note that PostgreSQL has an error on -ve values,
-- but MySQL just returns NULL.

-- TRUNCATE()
CREATE OR REPLACE FUNCTION truncate(numeric, integer)
RETURNS numeric AS $$
  SELECT pg_catalog.trunc($1, $2)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- INET_ATON()
-- Credit: Michael Fuhr
CREATE OR REPLACE FUNCTION inet_aton(text)
RETURNS bigint AS $$
  DECLARE
      a text[];
      b text[4];
      up int;
      family int;
      i int;
  BEGIN
      IF position(':' in $1) > 0 THEN
        family = 6;
      ELSE
        family = 4;
      END IF;
      -- mysql doesn't support IPv6 yet, it seems
      IF family = 6 THEN
        RETURN NULL;
      END IF;
      a = pg_catalog.string_to_array($1, '.');
      up = array_upper(a, 1);
      IF up = 4 THEN
        -- nothing to do
        b = a;
      ELSIF up = 3 THEN
        -- 127.1.2 = 127.1.0.2
        b = array[a[1], a[2], '0', a[3]];
      ELSIF up = 2 THEN
        -- 127.1 = 127.0.0.1
        b = array[a[1], '0', '0', a[2]];
      ELSIF up = 1 THEN
        -- 127 = 0.0.0.127
        b = array['0', '0', '0', a[1]];
      END IF;
      i = 1;
      -- handle 127..1
      WHILE i <= 4 LOOP
        IF length(b[i]) = 0 THEN
          b[i] = '0';
        END IF;
        i = i + 1;
      END LOOP;
      RETURN (b[1]::bigint << 24) | (b[2]::bigint << 16) |
              (b[3]::bigint << 8) | b[4]::bigint;
  END
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- INET_NTOA()
-- done in SQL to take advantage of inlining
CREATE OR REPLACE FUNCTION inet_ntoa(bigint)
RETURNS text AS $$
SELECT CASE WHEN $1 > 4294967295 THEN NULL ELSE
    ((($1::bigint >> 24) % 256) + 256) % 256 operator(pg_catalog.||) '.' operator(pg_catalog.||)
    ((($1::bigint >> 16) % 256) + 256) % 256 operator(pg_catalog.||) '.' operator(pg_catalog.||)
    ((($1::bigint >>  8) % 256) + 256) % 256 operator(pg_catalog.||) '.' operator(pg_catalog.||)
    ((($1::bigint      ) % 256) + 256) % 256 END;
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- SLEEP()
CREATE OR REPLACE FUNCTION sleep(float)
RETURNS integer AS $$
  BEGIN
    IF $1 > 0 THEN
      WHILE pg_catalog.timeofday()::timestamp < (current_timestamp + interval '1 second' * $1) LOOP
      -- Do nothing
      END LOOP;
    END IF;
    RETURN 0;
  END
$$ STRICT VOLATILE LANGUAGE PLPGSQL;
-- ADDDATE()
-- Note: passing in the interval is different
CREATE OR REPLACE FUNCTION adddate(timestamp without time zone, interval)
RETURNS timestamp without time zone AS $$
    SELECT $1 + $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION adddate(timestamp without time zone, integer)
RETURNS timestamp without time zone AS $$
  SELECT $1 + (INTERVAL '1 day' * $2)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- ADDTIME()
-- Note: requires casting if you install both versions
CREATE OR REPLACE FUNCTION addtime(timestamp without time zone, interval)
RETURNS timestamp without time zone AS $$
  SELECT $1 + $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION addtime(interval, interval)
RETURNS interval AS $$
  SELECT $1 + $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- CONVERT_TZ()
CREATE OR REPLACE FUNCTION convert_tz(timestamp without time zone, text, text)
RETURNS timestamp without time zone AS $$
  SELECT CASE
    WHEN POSITION(':' IN $3) = 0 THEN
      ($1 operator(pg_catalog.||) ' ' operator(pg_catalog.||) $2)::timestamp with time zone AT TIME ZONE $3
    ELSE
      ($1 operator(pg_catalog.||) ' ' operator(pg_catalog.||) $2)::timestamp with time zone AT TIME ZONE $3::interval
    END
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- CURDATE()
CREATE OR REPLACE FUNCTION curdate()
RETURNS date AS $$
  SELECT CURRENT_DATE
$$ VOLATILE LANGUAGE SQL;

-- CURTIME()
CREATE OR REPLACE FUNCTION curtime()
RETURNS time without time zone AS $$
  SELECT LOCALTIME(0)
$$ VOLATILE LANGUAGE SQL;

-- DATEDIFF()
CREATE OR REPLACE FUNCTION datediff(date, date)
RETURNS integer AS $$
  SELECT $1 - $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- DATE_ADD()
CREATE OR REPLACE FUNCTION date_add(timestamp without time zone, interval)
RETURNS timestamp without time zone AS $$
  SELECT $1 + $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- DATE_FORMAT()
CREATE OR REPLACE FUNCTION date_format(timestamp without time zone, text)
RETURNS text AS $$
  DECLARE
    i int := 1;
    temp text := '';
    c text;
    n text;
    res text;
  BEGIN
    WHILE i <= pg_catalog.length($2) LOOP
      -- Look at current character
      c := SUBSTRING ($2 FROM i FOR 1);
      -- If it's a '%' and not the last character then process it as a placeholder
      IF c = '%' AND i != pg_catalog.length($2) THEN
        n := SUBSTRING ($2 FROM (i + 1) FOR 1);
        SELECT INTO res CASE
          WHEN n = 'a' THEN pg_catalog.to_char($1, 'Dy') 
          WHEN n = 'b' THEN pg_catalog.to_char($1, 'Mon') 
          WHEN n = 'c' THEN pg_catalog.to_char($1, 'FMMM') 
          WHEN n = 'D' THEN pg_catalog.to_char($1, 'FMDDth') 
          WHEN n = 'd' THEN pg_catalog.to_char($1, 'DD') 
          WHEN n = 'e' THEN pg_catalog.to_char($1, 'FMDD') 
          WHEN n = 'f' THEN pg_catalog.to_char($1, 'US') 
          WHEN n = 'H' THEN pg_catalog.to_char($1, 'HH24') 
          WHEN n = 'h' THEN pg_catalog.to_char($1, 'HH12') 
          WHEN n = 'I' THEN pg_catalog.to_char($1, 'HH12')      
          WHEN n = 'i' THEN pg_catalog.to_char($1, 'MI') 
          WHEN n = 'j' THEN pg_catalog.to_char($1, 'DDD') 
          WHEN n = 'k' THEN pg_catalog.to_char($1, 'FMHH24') 
          WHEN n = 'l' THEN pg_catalog.to_char($1, 'FMHH12') 
          WHEN n = 'M' THEN pg_catalog.to_char($1, 'FMMonth') 
          WHEN n = 'm' THEN pg_catalog.to_char($1, 'MM') 
          WHEN n = 'p' THEN pg_catalog.to_char($1, 'AM') 
          WHEN n = 'r' THEN pg_catalog.to_char($1, 'HH12:MI:SS AM') 
          WHEN n = 'S' THEN pg_catalog.to_char($1, 'SS')
          WHEN n = 's' THEN pg_catalog.to_char($1, 'SS')
          WHEN n = 'T' THEN pg_catalog.to_char($1, 'HH24:MI:SS') 
          WHEN n = 'U' THEN pg_catalog.lpad(week($1::date, 0)::text, 2, '0')
          WHEN n = 'u' THEN pg_catalog.lpad(week($1::date, 1)::text, 2, '0')
          WHEN n = 'V' THEN pg_catalog.lpad(week($1::date, 2)::text, 2, '0')
          WHEN n = 'v' THEN pg_catalog.lpad(week($1::date, 3)::text, 2, '0')
          WHEN n = 'W' THEN pg_catalog.to_char($1, 'FMDay') 
          WHEN n = 'w' THEN EXTRACT(DOW FROM $1)::text
          WHEN n = 'X' THEN pg_catalog.lpad(((_calc_week($1::date, _week_mode(2)))[2])::text, 4, '0')
          WHEN n = 'x' THEN pg_catalog.lpad(((_calc_week($1::date, _week_mode(3)))[2])::text, 4, '0')
          WHEN n = 'Y' THEN pg_catalog.to_char($1, 'YYYY') 
          WHEN n = 'y' THEN pg_catalog.to_char($1, 'YY')
          WHEN n = '%' THEN pg_catalog.to_char($1, '%')
          ELSE NULL
        END;
        temp := temp operator(pg_catalog.||) res;
        i := i + 2;
      ELSE
        -- Otherwise just append the character to the string
        temp = temp operator(pg_catalog.||) c;
        i := i + 1;
      END IF; 
    END LOOP;
    RETURN temp;
  END
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- DATE_SUB()
CREATE OR REPLACE FUNCTION date_sub(timestamp without time zone, interval)
RETURNS timestamp without time zone AS $$
  SELECT $1 - $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- DAY()
CREATE OR REPLACE FUNCTION day(date)
RETURNS integer AS $$
  SELECT EXTRACT(DAY FROM DATE($1))::integer
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- DAYNAME()
CREATE OR REPLACE FUNCTION dayname(date)
RETURNS text AS $$
  SELECT pg_catalog.to_char($1, 'FMDay')
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- FROM_DAYS()
CREATE OR REPLACE FUNCTION from_days(integer)
RETURNS date AS $$
  SELECT ('0001-01-01 BC'::date + $1 * INTERVAL '1 day')::date
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- FROM_UNIXTIME()
-- Returns local time?  Is this actually the same as MySQL?
-- Depends on: DATE_FORMAT()
CREATE OR REPLACE FUNCTION from_unixtime(bigint)
RETURNS timestamp without time zone AS $$
  SELECT pg_catalog.to_timestamp($1)::timestamp without time zone
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION from_unixtime(bigint, text)
RETURNS text AS $$
  SELECT date_format(from_unixtime($1), $2)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- GET_FORMAT()
-- Note that first parameter needs to be quoted in this version
CREATE OR REPLACE FUNCTION get_format(text, text)
RETURNS text AS $$
  SELECT CASE
    WHEN $1 ILIKE 'DATE' THEN
      CASE WHEN $2 ILIKE 'USA' THEN '%m.%d.%Y'
      WHEN $2 ILIKE 'JIS' OR $2 ILIKE 'ISO' THEN '%Y-%m-%d'
      WHEN $2 ILIKE 'EUR' THEN '%d.%m.%Y'
      WHEN $2 ILIKE 'INTERNAL' THEN '%Y%m%d'
      ELSE NULL
      END
    WHEN $1 ILIKE 'DATETIME' THEN
      CASE WHEN $2 ILIKE 'USA' OR $2 ILIKE 'EUR' THEN '%Y-%m-%d-%H.%i.%s'
      WHEN $2 ILIKE 'JIS' OR $2 ILIKE 'ISO' THEN '%Y-%m-%d %H:%i:%s'
      WHEN $2 ILIKE 'INTERNAL' THEN '%Y%m%d%H%i%s'
      ELSE NULL
      END
    WHEN $1 ILIKE 'TIME' THEN
      CASE WHEN $2 ILIKE 'USA' THEN '%h:%i:%s %p'
      WHEN $2 ILIKE 'JIS' OR $2 ILIKE 'ISO' THEN '%H:%i:%s'
      WHEN $2 ILIKE 'EUR' THEN 'H.%i.%S'
      WHEN $2 ILIKE 'INTERNAL' THEN '%H%i%s'
      ELSE NULL
      END
    ELSE
      NULL
    END
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- LAST_DAY()
-- Note that for illegal timestamps this function raises an error,
-- whereas under MySQL it returns NULL
CREATE OR REPLACE FUNCTION last_day(timestamp)
RETURNS date AS $$ 
  SELECT CASE
    WHEN EXTRACT(MONTH FROM $1) = 12 THEN
      (((EXTRACT(YEAR FROM $1) + 1) operator(pg_catalog.||) '-01-01')::date - INTERVAL '1 day')::date
    ELSE
      ((EXTRACT(YEAR FROM $1) operator(pg_catalog.||) '-' operator(pg_catalog.||) (EXTRACT(MONTH FROM $1) + 1) operator(pg_catalog.||) '-01')::date - INTERVAL '1 day')::date
    END
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- MAKEDATE()
CREATE OR REPLACE FUNCTION makedate(integer, integer)
RETURNS date AS $$
  SELECT CASE WHEN $2 > 0 THEN
    (($1 operator(pg_catalog.||) '-01-01')::date + ($2 - 1) * INTERVAL '1 day')::date
  ELSE
    NULL
  END
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- MAKETIME()
CREATE OR REPLACE FUNCTION maketime(integer, integer, integer)
RETURNS interval AS $$
  SELECT ($1 operator(pg_catalog.||) ':' operator(pg_catalog.||) $2 operator(pg_catalog.||) ':' operator(pg_catalog.||) $3)::interval
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- MONTHNAME()
CREATE OR REPLACE FUNCTION monthname(date)
RETURNS text AS $$
  SELECT pg_catalog.to_char($1, 'FMMonth')
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- NOW()
-- Part of base PostgreSQL

-- PERIOD_ADD()
CREATE OR REPLACE FUNCTION period_add(integer, integer)
RETURNS text AS $$
  DECLARE
    period text;
    base date;
    baseyear integer;
  BEGIN
      IF pg_catalog.length($1) < 4 THEN
        period := pg_catalog.lpad($1, 4, 0);
      ELSIF pg_catalog.length($1) = 5 THEN
        period := pg_catalog.lpad($1, 6, 0);
      ELSE
        period := $1;
      END IF;

      IF pg_catalog.length(period) = 4 THEN
        baseyear := SUBSTRING(period FROM 1 FOR 2);
        IF baseyear BETWEEN 70 AND 99 THEN
          baseyear := baseyear + 1900;
        ELSE
          baseyear := baseyear + 2000;
        END IF;
        base := (baseyear operator(pg_catalog.||) '-' operator(pg_catalog.||) SUBSTRING(period FROM 3) operator(pg_catalog.||) '-01')::date;
      ELSIF pg_catalog.length(period) = 6 THEN
        base := (SUBSTRING(period FROM 1 FOR 4) operator(pg_catalog.||) '-' operator(pg_catalog.||) SUBSTRING(period FROM 5) operator(pg_catalog.||) '-01')::date;
      ELSE
        RETURN NULL;
      END IF;

      base := base + (INTERVAL '1 month' * $2);
      RETURN pg_catalog.lpad(EXTRACT(YEAR FROM base), 4, '0') operator(pg_catalog.||) pg_catalog.lpad(EXTRACT(MONTH FROM base), 2, '0');
  END
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- PERIOD_DIFF()
CREATE OR REPLACE FUNCTION period_diff(integer, integer)
RETURNS integer AS $$
  DECLARE
    baseyear integer;
    period1 text;
    period2 text;
    months1 integer;
    months2 integer;
  BEGIN
      IF pg_catalog.length($1) < 4 THEN
        period1 := pg_catalog.lpad($1, 4, 0);
      ELSIF pg_catalog.length($1) = 5 THEN
        period1 := pg_catalog.lpad($1, 6, 0);
      ELSE
        period1 := $1;
      END IF;

      IF pg_catalog.length(period1) = 4 THEN
        baseyear := SUBSTRING(period1 FROM 1 FOR 2);
        IF baseyear BETWEEN 70 AND 99 THEN
          baseyear := baseyear + 1900;
        ELSE
          baseyear := baseyear + 2000;
        END IF;
        months1 := baseyear * 12 + SUBSTRING(period1 FROM 3)::integer;
      ELSIF pg_catalog.length(period1) = 6 THEN
        months1 := SUBSTRING(period1 FROM 1 FOR 4)::integer * 12 + SUBSTRING(period1 FROM 5)::integer;
      ELSE
        RETURN NULL;
      END IF;

      IF pg_catalog.length($2) < 4 THEN
        period2 := pg_catalog.lpad($2, 4, 0);
      ELSIF pg_catalog.length($2) = 5 THEN
        period2 := pg_catalog.lpad($2, 6, 0);
      ELSE
        period2 := $2;
      END IF;

      IF pg_catalog.length(period2) = 4 THEN
        baseyear := SUBSTRING(period2 FROM 1 FOR 2);
        IF baseyear BETWEEN 70 AND 99 THEN
          baseyear := baseyear + 1900;
        ELSE
          baseyear := baseyear + 2000;
        END IF;
        months2 := baseyear * 12 + SUBSTRING(period2 FROM 3)::integer;
      ELSIF pg_catalog.length(period2) = 6 THEN
        months2 := SUBSTRING(period2 FROM 1 FOR 4)::integer * 12 + SUBSTRING(period2 FROM 5)::integer;
      ELSE
        RETURN NULL;
      END IF;

      RETURN months1 - months2;
  END
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- SEC_TO_TIME()
CREATE OR REPLACE FUNCTION sec_to_time(bigint)
RETURNS interval AS $$
  SELECT $1 * INTERVAL '1 second'
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- STR_TO_DATE()
-- Note: Doesn't handle weeks of years yet and will return different results
-- to MySQL if you pass in an invalid timestamp
CREATE OR REPLACE FUNCTION str_to_date(text, text)
RETURNS timestamp without time zone AS $$
  DECLARE
    i int := 1;
    temp text := '';
    c text;
    n text; res text;
  BEGIN
    WHILE i <= pg_catalog.length($2) LOOP
      -- Look at current character
      c := SUBSTRING ($2 FROM i FOR 1);
      -- If it's a '%' and not the last character then process it as a placeholder
      IF c = '%' AND i != pg_catalog.length($2) THEN
        n := SUBSTRING ($2 FROM (i + 1) FOR 1);
        SELECT INTO res CASE
          WHEN n = 'a' THEN 'Dy' 
          WHEN n = 'b' THEN 'Mon' 
          WHEN n = 'c' THEN 'FMMM' 
          WHEN n = 'D' THEN 'FMDDth' 
          WHEN n = 'd' THEN 'DD' 
          WHEN n = 'e' THEN 'FMDD' 
          WHEN n = 'f' THEN 'US' 
          WHEN n = 'H' THEN 'HH24' 
          WHEN n = 'h' THEN 'HH12' 
          WHEN n = 'I' THEN 'HH12'      
          WHEN n = 'i' THEN 'MI' 
          WHEN n = 'j' THEN 'DDD' 
          WHEN n = 'k' THEN 'FMHH24' 
          WHEN n = 'l' THEN 'FMHH12' 
          WHEN n = 'M' THEN 'FMMonth' 
          WHEN n = 'm' THEN 'MM' 
          WHEN n = 'p' THEN 'AM' 
          WHEN n = 'r' THEN 'HH12:MI:SS AM' 
          WHEN n = 'S' THEN 'SS'
          WHEN n = 's' THEN 'SS'
          WHEN n = 'T' THEN 'HH24:MI:SS' 
          WHEN n = 'U' THEN '?'
          WHEN n = 'u' THEN '?'
          WHEN n = 'V' THEN '?'
          WHEN n = 'v' THEN '?'
          WHEN n = 'W' THEN 'FMDay' 
          WHEN n = 'w' THEN '?'
          WHEN n = 'X' THEN '?' 
          WHEN n = 'x' THEN '?' 
          WHEN n = 'Y' THEN 'YYYY' 
          WHEN n = 'y' THEN 'YY'
          WHEN n = '%' THEN '%'
          ELSE NULL
        END;
        temp := temp operator(pg_catalog.||) res;
        i := i + 2;
      ELSE
        -- Otherwise just append the character to the string
        temp = temp operator(pg_catalog.||) c;
        i := i + 1;
      END IF; 
    END LOOP;
    RETURN pg_catalog.to_timestamp($1, temp)::timestamp without time zone;
  END
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- SUBDATE()
-- Note: passing in the interval is different
CREATE OR REPLACE FUNCTION subdate(timestamp without time zone, interval)
RETURNS timestamp without time zone AS $$
    SELECT $1 - $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION subdate(timestamp without time zone, integer)
RETURNS timestamp without time zone AS $$
  SELECT $1 - (INTERVAL '1 day' * $2)
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- SUBTIME
-- Note: requires casting if you install both versions
CREATE OR REPLACE FUNCTION subtime(timestamp without time zone, interval)
RETURNS timestamp without time zone AS $$
  SELECT $1 - $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION subtime(interval, interval)
RETURNS interval AS $$
  SELECT $1 - $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- SYSDATE()
CREATE OR REPLACE FUNCTION sysdate()
RETURNS timestamp without time zone AS $$
  SELECT pg_catalog.timeofday()::timestamp(0) without time zone
$$ VOLATILE LANGUAGE SQL;

-- TIME()
-- Not possible to implement

-- TIMEDIFF()
-- Note: requires casting if you install both versions
CREATE OR REPLACE FUNCTION timediff(timestamp without time zone, timestamp without time zone)
RETURNS interval AS $$
  SELECT $1 - $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OR REPLACE FUNCTION timediff(time without time zone, time without time zone)
RETURNS interval AS $$
  SELECT $1 - $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- TIMESTAMP()
-- Not possible to implement

-- TIMESTAMPADD()
-- Note that first parameter needs to be quoted in this version
CREATE OR REPLACE FUNCTION timestampadd(text, integer, timestamp without time zone)
RETURNS timestamp without time zone AS $$
  SELECT $3 + ($2 operator(pg_catalog.||) ' ' operator(pg_catalog.||) $1)::interval
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- TIMESTAMPDIFF()
-- Note that first parameter needs to be quoted in this version

-- TIME_FORMAT()
CREATE OR REPLACE FUNCTION time_format(interval, text)
RETURNS text AS $$
  DECLARE
    i int := 1;
    temp text := '';
    c text;
    n text;
    res text;
  BEGIN
    WHILE i <= pg_catalog.length($2) LOOP
      -- Look at current character
      c := SUBSTRING ($2 FROM i FOR 1);
      -- If it's a '%' and not the last character then process it as a placeholder
      IF c = '%' AND i != pg_catalog.length($2) THEN
        n := SUBSTRING ($2 FROM (i + 1) FOR 1);
        SELECT INTO res CASE
          WHEN n = 'a' THEN '0'
          WHEN n = 'b' THEN '0'
          WHEN n = 'c' THEN '0'
          WHEN n = 'D' THEN '0'
          WHEN n = 'd' THEN '0'
          WHEN n = 'e' THEN '0'
          WHEN n = 'f' THEN pg_catalog.to_char($1, 'US') 
          WHEN n = 'H' THEN pg_catalog.to_char($1, 'HH24') 
          WHEN n = 'h' THEN pg_catalog.lpad(pg_catalog.to_char($1, 'HH12')::integer % 12, 2, '0')
          WHEN n = 'I' THEN pg_catalog.lpad(pg_catalog.to_char($1, 'HH12')::integer % 12, 2, '0')
          WHEN n = 'i' THEN pg_catalog.to_char($1, 'MI') 
          WHEN n = 'j' THEN '0'
          WHEN n = 'k' THEN pg_catalog.to_char($1, 'FMHH24') 
          WHEN n = 'l' THEN (pg_catalog.to_char($1, 'FMHH12')::integer % 12)::text
          WHEN n = 'M' THEN '0'
          WHEN n = 'm' THEN '0'
          WHEN n = 'p' THEN pg_catalog.to_char($1, 'AM') 
          WHEN n = 'r' THEN pg_catalog.lpad(pg_catalog.to_char($1, 'HH12')::integer % 12, 2, '0') 
                              operator(pg_catalog.||) 
                              pg_catalog.to_char($1, ':MI:SS ') 
                              operator(pg_catalog.||) 
                              CASE WHEN pg_catalog.to_char($1, 'FMHH24')::integer <= 11 THEN 'AM' ELSE 'PM' END
          WHEN n = 'S' THEN pg_catalog.to_char($1, 'SS')
          WHEN n = 's' THEN pg_catalog.to_char($1, 'SS')
          WHEN n = 'T' THEN pg_catalog.to_char($1, 'HH24:MI:SS') 
          WHEN n = 'U' THEN '0'
          WHEN n = 'u' THEN '0'
          WHEN n = 'V' THEN '0'
          WHEN n = 'v' THEN '0'
          WHEN n = 'W' THEN '0'
          WHEN n = 'w' THEN '0'
          WHEN n = 'X' THEN '0'
          WHEN n = 'x' THEN '0'
          WHEN n = 'Y' THEN '0'
          WHEN n = 'y' THEN '0'
          WHEN n = '%' THEN pg_catalog.to_char($1, '%')
          ELSE NULL
        END;
        temp := temp operator(pg_catalog.||) res;
        i := i + 2;
      ELSE
        -- Otherwise just append the character to the string
        temp = temp operator(pg_catalog.||) c;
        i := i + 1;
      END IF; 
    END LOOP;
    RETURN temp;
  END
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- TIME_TO_SEC()
CREATE OR REPLACE FUNCTION time_to_sec(interval)
RETURNS bigint AS $$
  SELECT (EXTRACT(HOURS FROM $1) * 3600
    + EXTRACT(MINUTES FROM $1) * 60
    + EXTRACT(SECONDS FROM $1))::bigint
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- TO_DAYS()
-- XXX: Haven't done integer variant
CREATE OR REPLACE FUNCTION to_days(date)
RETURNS integer AS $$
  SELECT $1 - '0001-01-01 BC'::date
$$ IMMUTABLE STRICT LANGUAGE SQL;

-- UNIX_TIMESTAMP()
CREATE OR REPLACE FUNCTION unix_timestamp()
RETURNS bigint AS $$
  SELECT EXTRACT(EPOCH FROM LOCALTIMESTAMP)::bigint
$$ VOLATILE LANGUAGE SQL;

-- XXX: This gives wrong answers? Time zones?
CREATE OR REPLACE FUNCTION unix_timestamp(timestamp without time zone)
RETURNS bigint AS $$
  SELECT EXTRACT(EPOCH FROM $1)::bigint
$$ VOLATILE LANGUAGE SQL;

-- UTC_DATE()
CREATE OR REPLACE FUNCTION utc_date()
RETURNS date AS $$
  SELECT (CURRENT_TIMESTAMP AT TIME ZONE 'UTC')::date
$$ VOLATILE LANGUAGE SQL;

-- UTC_TIME()
CREATE OR REPLACE FUNCTION utc_time()
RETURNS time(0) AS $$
  SELECT (CURRENT_TIMESTAMP AT TIME ZONE 'UTC')::time(0)
$$ VOLATILE LANGUAGE SQL;

-- UTC_TIMESTAMP()
CREATE OR REPLACE FUNCTION utc_timestamp()
RETURNS timestamp(0) AS $$
  SELECT (CURRENT_TIMESTAMP AT TIME ZONE 'UTC')::timestamp(0)
$$ VOLATILE LANGUAGE SQL;

-- WEEK()
CREATE OR REPLACE FUNCTION _week_mode(mode integer)
RETURNS integer AS $$
  DECLARE
    _WEEK_MONDAY_FIRST  CONSTANT integer := 1;
    _WEEK_FIRST_WEEKDAY CONSTANT integer := 4;
    week_format integer := mode & 7;
  BEGIN
    IF (week_format & _WEEK_MONDAY_FIRST) = 0 THEN
      week_format := week_format # _WEEK_FIRST_WEEKDAY;
    END IF;

    RETURN week_format;
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION _calc_weekday(qdate date, sundayfirst boolean)
RETURNS integer AS $$
  BEGIN
    RETURN (EXTRACT(DOW FROM qdate)::integer + CASE WHEN sundayfirst THEN 0 ELSE 6 END) % 7;
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION _calc_days_in_year(year integer)
RETURNS integer AS $$
  BEGIN
    IF (year & 3) = 0 AND ((year % 100) <> 0 OR (year % 400) = 0 AND year <> 0) THEN
      RETURN 366;
    ELSE
      RETURN 365;
    END IF;
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION _calc_week(qdate anyelement, behavior integer)
RETURNS integer[] AS $$
  DECLARE
    _WEEK_MONDAY_FIRST  CONSTANT integer := 1;
    _WEEK_YEAR          CONSTANT integer := 2;
    _WEEK_FIRST_WEEKDAY CONSTANT integer := 4;
    qyear         integer := EXTRACT(YEAR FROM qdate);
    qmonth        integer := EXTRACT(MONTH FROM qdate);
    qday          integer := EXTRACT(DAY FROM qdate);
    daynr         integer := EXTRACT(DOY FROM qdate);
    yday1         date    := pg_catalog.date_trunc('year', qdate);
    first_daynr   integer := 1;
    monday_first  boolean := (behavior & _WEEK_MONDAY_FIRST) <> 0;
    week_year     boolean := (behavior & _WEEK_YEAR) <> 0;
    first_weekday boolean := (behavior & _WEEK_FIRST_WEEKDAY) <> 0;
    weekday       integer := _calc_weekday(yday1, NOT monday_first);
    days          integer;
  BEGIN
    IF qmonth = 1 AND qday <= 7 - weekday THEN
      IF (NOT week_year) AND ((first_weekday AND weekday <> 0) OR (NOT first_weekday AND weekday >= 4)) THEN
        RETURN array[0, qyear];
      END IF;

      week_year := true;
      qyear := qyear - 1;
      days := _calc_days_in_year(qyear);
      first_daynr := first_daynr - days;
      weekday := (weekday + 53 * 7 - days) % 7;
    END IF;

    IF (first_weekday AND weekday <> 0) OR (NOT first_weekday AND weekday >= 4) THEN
      days := daynr - (first_daynr + (7 - weekday));
    ELSE
      days := daynr - (first_daynr - weekday);
    END IF;

    IF week_year AND days >= 52 * 7 THEN
      weekday := (weekday + _calc_days_in_year(qyear)) % 7;
      IF (NOT first_weekday AND weekday < 4) OR (first_weekday AND weekday = 0) THEN
	qyear := qyear + 1;
        RETURN array[1, qyear];
      END IF;
    END IF;

    RETURN array[days / 7 + 1, qyear];
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- YEARWEEK()
CREATE OR REPLACE FUNCTION yearweek(qdate date, mode integer)
RETURNS integer AS $$
  DECLARE
    _WEEK_YEAR CONSTANT integer := 2;
    a integer[] := _calc_week(qdate, _week_mode(mode | _WEEK_YEAR));
    week integer := a[1];
    year integer := a[2];
  BEGIN
    RETURN week + year * 100;
  END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION yearweek(date)
RETURNS integer AS $$
  SELECT yearweek($1, 0);
$$ IMMUTABLE STRICT LANGUAGE SQL;

--
-- is_datetime: used to determine if value passed to polymorphic 
-- function is a valid date/time value
--

CREATE OR REPLACE FUNCTION is_datetime ( anyelement )
RETURNS BOOLEAN AS $$
DECLARE d date;
        t TIMESTAMP;
        tz TIMESTAMPTZ;
BEGIN
   d = $1::DATE;
   RETURN TRUE;
EXCEPTION WHEN others THEN
        BEGIN
           t = $1::TIMESTAMP;
                RETURN TRUE;
        EXCEPTION WHEN others THEN
                BEGIN
                        tz = $1::TIMESTAMPTZ;
                        RETURN TRUE;
                EXCEPTION WHEN others THEN
                   RETURN FALSE;
                END;
        END;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

--
-- Polymorphic functions to allow one function 
-- to handle date / timestamp / timestamptz types
--

-- WEEKDAY()
CREATE OR REPLACE FUNCTION weekday( anyelement )
RETURNS integer AS $$
BEGIN
        IF is_datetime ( $1 ) THEN
                CASE WHEN EXTRACT(DOW FROM $1)::integer = 0 THEN
                        RETURN 6;
                     ELSE
                        RETURN EXTRACT(DOW FROM $1)::integer - 1;
                END CASE;
        END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- YEAR()
CREATE OR REPLACE FUNCTION year( anyelement )
RETURNS integer AS $$
BEGIN
        IF is_datetime ( $1 ) THEN
                RETURN EXTRACT(YEAR FROM $1)::integer;
        END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- MONTH()
CREATE OR REPLACE FUNCTION month( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN EXTRACT(MONTH FROM DATE($1))::integer;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- DAYOFMONTH()
CREATE OR REPLACE FUNCTION dayofmonth( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN EXTRACT(DAY FROM DATE($1))::integer;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- DAYOFWEEK()
CREATE OR REPLACE FUNCTION dayofweek( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN EXTRACT(DOW FROM DATE($1))::integer + 1;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- DAYOFYEAR()
CREATE OR REPLACE FUNCTION dayofyear( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN EXTRACT(DOY FROM DATE($1))::integer;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- HOUR()
CREATE OR REPLACE FUNCTION hour( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN EXTRACT (HOUR FROM $1)::integer;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- MICROSECOND()
CREATE OR REPLACE FUNCTION microsecond( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN (EXTRACT(MICROSECONDS FROM $1))::integer % 1000000;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- MINUTE()
CREATE OR REPLACE FUNCTION minute( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN EXTRACT(MINUTES FROM $1)::integer;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- QUARTER()
CREATE OR REPLACE FUNCTION quarter( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN EXTRACT(QUARTER FROM DATE($1))::integer;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- SECOND()
CREATE OR REPLACE FUNCTION second( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN EXTRACT(SECONDS FROM $1)::integer;
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION week( anyelement, integer )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN (_calc_week($1, _week_mode($2)))[1];
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

CREATE OR REPLACE FUNCTION week( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN week($1, 0);
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;

-- WEEKOFYEAR()
CREATE OR REPLACE FUNCTION weekofyear( anyelement )
RETURNS integer AS $$
BEGIN
	IF is_datetime ( $1 ) THEN
		RETURN week($1, 3);
	END IF;
	RAISE EXCEPTION 'Invalid date / time value --> %', $1;
END;
$$ IMMUTABLE STRICT LANGUAGE PLPGSQL;
-- <=> NULL SAFE COMPARISON
-- Note: needs casts in some circumstances
CREATE OR REPLACE FUNCTION _null_safe_cmp(anyelement, anyelement)
RETURNS boolean AS '
  SELECT NOT ($1 IS DISTINCT FROM $2)
' IMMUTABLE LANGUAGE SQL;

CREATE OPERATOR <=> (
  PROCEDURE = _null_safe_cmp,
  LEFTARG = anyelement,
  RIGHTARG = anyelement
);

-- &&
-- XXX: MySQL version has wacky null behaviour 
CREATE FUNCTION _and(boolean, boolean)
RETURNS boolean AS $$
  SELECT $1 AND $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OPERATOR && (
  leftarg = boolean,
  rightarg = boolean,
  procedure = _and,
  commutator = &&
);

-- ||
-- XXX: MySQL version has wacky null behaviour 
-- This replaces the SQL standard || concatenation operator
CREATE FUNCTION _or(boolean, boolean)
RETURNS boolean AS $$
  SELECT $1 OR $2
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OPERATOR || (
  leftarg = boolean,
  rightarg = boolean,
  procedure = _or,
  commutator = || 
);
